/** FuzzyTree.java
 *  Copyright (c) 2001 Donna Malayeri
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 2
 *  of the License, or (at your option) any later version.
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

package edu.umd.cs.fuzzyTree;

import javax.swing.*;
import javax.swing.tree.*;
import javax.swing.event.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;

/**
 *  Component which displays a partial tree of an XML document. A limited number of nodes
 *  of the entire tree are displayed, and the FuzzyTree attempts to show the "best" nodes
 *  for the given tree. The best node is based on a variety of parameters and each node is
 *  given a score. See {@link ScoreGenerator}.  This class intercepts all regular JTree
 *  user expand and collapse events and vetos them. Expanding and collapsing are instead
 *  handled by a mouse listener on the tree (so either the node or its icon can be clicked
 *  on to expand/collapse.
 *
 *  <p> Many of the public methods in this class are simply delegates to methods in
 *  FuzzyTreeModel.
 *
 *  <p>A custom node renderer is used to display a node in one
 *  of three possible states: fully open (all of its children are displayed),
 *  partially open (at least one of its children are displayed), or
 *  closed (none of it children are displayed).  A node's parent may not be in the displayed
 *  tree; in such a case it is an "orphan" and is placed as the child of its closest ancestor.
 *  Such nodes have a different icon.
 *
 * <p>Images are used by the custom node renderer:
 *  <ul>
 *  <li>Closed node: <code>images/closed.gif</code>
 *  <li>Fully open node: <code>images/open.gif</code>
 *  <li>Partially open node: <code>images/partly_open.gif</code>
 *  <li>Leaf node: <code>images/clear.gif</code>
 *  </ul>
 *
 *  @see JTree
 *
 *  @author     Donna Malayeri
 *  @version    24-Jul-01
 */

public class FuzzyTree extends JTree implements TreeWillExpandListener
{
	/** Used to distinguish between programmatic expansions and user expansions. */
	private boolean userIsExpanding = true;
	/** The tree's tree model, same as that returned by <code>getModel()</code>. */
	FuzzyTreeModel _model;

	/** Default constructor, provided only so FuzzyTree will be a bean. If this constructor is used,
	 *  {@link #setXMLFile} should be called to set the tree model.
	 */
	public FuzzyTree()
	{
		setSelectionModel(null);     // no selections
		setShowsRootHandles(false);  // don't show root handles
		setRootVisible(true);
		addTreeWillExpandListener(this);      // listen for expand and collapse events

		if (getUI() instanceof BasicTreeUI)
		{
			BasicTreeUI basicUI = (BasicTreeUI) getUI();
			basicUI.setExpandedIcon(null);      // get rid of default expand/collapse handles
			basicUI.setCollapsedIcon(null);
		}
	}

	/** Create a FuzzyTree using the specified file as data.
	 *  @param xmlFilename      the string filename of the XML file
	 */
	public FuzzyTree(String xmlFilename) throws DataGraph.DataGraphException
	{
		this();
		setXMLFile(xmlFilename);
	}

	/** Create a new tree model using the specified file as data.
	 *  @param xmlFilename      the string filename of the XML file
	 */
	public void setXMLFile(String xmlFilename) throws DataGraph.DataGraphException
	{
		_model = new FuzzyTreeModel(xmlFilename);
		setModel(_model);

		setCellRenderer(new NodeRenderer());        // use a custom node renderer
		addMouseListener(new TreeMouseListener());  // listen for mouse clicks in the tree
	}

	/** Sets the number of displayed nodes in the tree to the new value.
	 *  @param  newValue        new number of nodes to display
	 */
	public void setNumDisplayedNodes(int newValue)
	{
		_model.setNumDisplayedNodes(newValue);
		showEntireTree();
	}

	/** Returns the number of nodes displayed in the tree.
	 *  @return     the number of nodes displayed in the tree
	 */
	public int getNumDisplayedNodes()
	{
		return _model.getNumDisplayedNodes();
	}

	/** Returns the total number of nodes in the tree.
	 *  @return the total number of nodes in the tree
	 */
	public int getTotalNodes()
	{
		return _model.getTotalNodes();
	}

	/** Get parameters used in score adjusting algorithms
	 *  @return parameters used by score adjusting algorithms */
	public FuzzyTreeParameters getParameters()
	{
		return _model.getParameters();
	}

	/** Set parameters for score adjusting algorithms
	 *  @param newParameters    new parameter values
	 *  */
	public void setParameters(FuzzyTreeParameters params)
	{
		_model.setParameters(params);
	}

	/** Traps all user expansions and disallows them. */
	public void treeWillExpand(TreeExpansionEvent e) throws ExpandVetoException
	{
		if (userIsExpanding)
			throw new ExpandVetoException(e);
	}

	/** Traps all user collapses and disallows them. */
	public void treeWillCollapse(TreeExpansionEvent e) throws ExpandVetoException
	{
		throw new ExpandVetoException(e);
	}

	/** Expands the entire displayed tree, but does not display nodes that are not part of
	 *  the display tree of the tree model.
	 */
	protected void showEntireTree()
	{
		userIsExpanding = false;

		int row = 0;
		while (row < getRowCount())
		{
			expandRow(row);
			row++;
		}

		userIsExpanding = true;
	}

	class TreeMouseListener extends MouseAdapter
	{
		/** Expand/collapse are implemented using the listener, instead of the JTree's handles.
		 *  <p>Shift+click: fully expands or collapses<p>
		 *  Partially open nodes are expanded by default;
		 *  control+click will always collapse non-closed nodes.
		 */
		public void mouseClicked(MouseEvent e)
		{
			// some mouse clicks are getting lost--not sure why. This method is not even getting
			// called in those cases.

			int selRow = getRowForLocation(e.getX(), e.getY());
			TreePath selPath = getPathForLocation(e.getX(), e.getY());

			if (selRow != -1)
			{
				Object node = selPath.getLastPathComponent();
				boolean fullyExpand = e.isShiftDown();

				if (!_model.isLeaf(node))
				{
					int nodeState = _model.getNodeState(node);

					if ( nodeState == FuzzyTreeModel.NODE_FULLY_OPEN ||
						(nodeState == FuzzyTreeModel.NODE_PARTIALLY_OPEN && e.isControlDown()))
						_model.userCollapseSubtree(node, fullyExpand);
					else
						_model.userExpandSubtree(node, fullyExpand);

					showEntireTree();
				}
			}
		}
	}

	/** Renderer to color nodes and set icons.
	 *  <p>Images:<br>
	 *  <ul>
	 *  <li>Closed node: <code>images/closed.gif</code>
	 *  <li>Fully open node: <code>images/open.gif</code>
	 *  <li>Partially open node: <code>images/partly_open.gif</code>
	 *  <li>Leaf node: <code>images/clear.gif</code>
	 *  </ul>
	 *  */
	class NodeRenderer extends DefaultTreeCellRenderer
	{
		ImageIcon fullyOpenIcon;
		ImageIcon orphanFullyOpenIcon;
		ImageIcon partlyOpenIcon;
		ImageIcon orphanPartlyOpenIcon;
		ImageIcon closedIcon;
		ImageIcon orphanClosedIcon;
		ImageIcon leafIcon;
		ImageIcon orphanLeafIcon;

		public NodeRenderer()
		{
			fullyOpenIcon           = new ImageIcon("images/open.gif");
			orphanFullyOpenIcon     = new ImageIcon("images/open_o.gif");

			partlyOpenIcon          = new ImageIcon("images/partly_open.gif");
			orphanPartlyOpenIcon    = new ImageIcon("images/partly_open_o.gif");

			closedIcon              = new ImageIcon("images/closed.gif");
			orphanClosedIcon        = new ImageIcon("images/closed_o.gif");

			leafIcon                = new ImageIcon("images/leaf.gif");
			orphanLeafIcon          = new ImageIcon("images/leaf_o.gif");
		}

		public Component getTreeCellRendererComponent(JTree tree, Object node,
							boolean sel, boolean expanded, boolean leaf,
							int row, boolean hasFocus)
		{

			sel = _model.isActiveNode(node);
			// old - orphan nodes were colored blue, now different icons are used
			// setTextNonSelectionColor(FuzzyTreeModel.isOrphan(node) ? Color.blue : Color.black);
			super.getTreeCellRendererComponent(tree, node, sel, expanded, leaf, row, hasFocus);

			boolean isOrphan = FuzzyTreeModel.isOrphan(node);
			int nodeState = _model.getNodeState(node);
			if (nodeState == FuzzyTreeModel.NODE_FULLY_OPEN)
				setIcon(isOrphan ? orphanFullyOpenIcon : fullyOpenIcon);
			else if (!leaf && nodeState == FuzzyTreeModel.NODE_CLOSED)
				setIcon(isOrphan ? orphanClosedIcon : closedIcon);
			else if (leaf)
				setIcon(isOrphan ? orphanLeafIcon : leafIcon);
			else
				setIcon(isOrphan ? orphanPartlyOpenIcon : partlyOpenIcon);

			return this;
		}
	}
}
